use std::fmt::Debug;
use std::hash::Hash;
use std::io::{BufRead, BufReader};
use std::process::Stdio;
use std::sync::{Arc, RwLock};
use std::thread;
use std::thread::sleep;
use std::time::Duration;

use log::{error, warn};
use once_cell::sync::Lazy;
use serde::{Deserialize, Serialize};
use tokio::sync::broadcast;
use tokio::sync::broadcast::error::RecvError;

use deckster_mode::shared::handler_communication::{HandlerCommand, HandlerEvent, HandlerInitializationError, InitialHandlerMessage, KeyEvent};
use deckster_mode::shared::path::KeyPath;
use deckster_mode::shared::state::KeyStyleByStateMap;
use deckster_mode::{send_command, DecksterHandler};

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
#[serde(tag = "mode", rename_all = "kebab-case")]
pub enum KeyConfig {
    Play(ButtonConfig),
    Pause(ButtonConfig),
    PlayPause(ButtonConfig),
    Previous(ButtonConfig),
    Next(ButtonConfig),
    Shuffle(ShuffleConfig),
    Loop(LoopConfig),
}

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct ButtonConfig {
    #[serde(default)]
    pub style: KeyStyleByStateMap<ButtonState>,
}

#[derive(Debug, Eq, PartialEq)]
pub enum ButtonCommand {
    PlayPause,
    Play,
    Pause,
    Previous,
    Next,
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum ButtonState {
    Inactive,
    Playing,
    Paused,
}

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct ShuffleConfig {
    #[serde(default)]
    pub style: KeyStyleByStateMap<ShuffleState>,
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum ShuffleState {
    Inactive,
    Off,
    On,
}

#[derive(Debug, Clone, PartialEq, Serialize, Deserialize)]
pub struct LoopConfig {
    #[serde(default)]
    pub style: KeyStyleByStateMap<LoopState>,
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Serialize, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum LoopState {
    Inactive,
    None,
    Single,
    All,
}

struct PlayerctlStateWatcher<S: Clone + Debug + Send + Sync + 'static> {
    state: Arc<RwLock<S>>,
    new_states: broadcast::Sender<S>,
}

impl<S: Clone + Debug + Send + Sync + 'static> PlayerctlStateWatcher<S> {
    pub fn new(initial_state: S, subcommand: &'static str, parse_state: impl 'static + Fn(&String) -> S + Send) -> Self {
        let state = Arc::new(RwLock::new(initial_state));
        let cloned_state = Arc::clone(&state);

        let new_states = broadcast::Sender::new(1);
        let cloned_new_states = new_states.clone();

        thread::spawn(move || {
            loop {
                let players = std::process::Command::new("playerctl")
                    .args(["-s", "-l"])
                    .stderr(Stdio::null())
                    .stdin(Stdio::null())
                    .output()
                    .unwrap();

                if players.stdout.iter().filter(|c| **c == b'\n').count() > 0 {
                    break;
                }

                sleep(Duration::from_secs(1));
            }

            let command = std::process::Command::new("playerctl")
                .args(["-s", "-F", subcommand])
                .stdout(Stdio::piped())
                .stderr(Stdio::null())
                .stdin(Stdio::null())
                .spawn()
                .unwrap();

            let stdout = BufReader::new(command.stdout.unwrap());

            for line in stdout.lines() {
                let line = line.unwrap();
                let new_state = parse_state(&line);

                {
                    let mut g = cloned_state.write().unwrap();
                    *g = new_state.clone();
                    if cloned_new_states.send(new_state).is_err() {
                        log::warn!("No one is listening to player state changes")
                    }
                }
            }
        });

        PlayerctlStateWatcher { state, new_states }
    }

    fn subscribe_to_state(&self) -> broadcast::Receiver<S> {
        let r = self.new_states.subscribe();
        self.new_states.send(self.state.read().unwrap().clone()).unwrap();
        r
    }
}

static STATE_WATCHER_PLAYING: Lazy<PlayerctlStateWatcher<ButtonState>> = Lazy::new(|| {
    PlayerctlStateWatcher::new(ButtonState::Inactive, "status", |s| match s.as_ref() {
        "Playing" => ButtonState::Playing,
        "Paused" => ButtonState::Paused,
        "" | "Stopped" => ButtonState::Inactive,
        _ => panic!("Unknown state: {}", s),
    })
});

static STATE_WATCHER_SHUFFLE: Lazy<PlayerctlStateWatcher<ShuffleState>> = Lazy::new(|| {
    PlayerctlStateWatcher::new(ShuffleState::Inactive, "shuffle", |s| match s.as_ref() {
        "Off" => ShuffleState::Off,
        "On" => ShuffleState::On,
        "" => ShuffleState::Inactive,
        _ => panic!("Unknown state: {}", s),
    })
});

static STATE_WATCHER_LOOP: Lazy<PlayerctlStateWatcher<LoopState>> = Lazy::new(|| {
    PlayerctlStateWatcher::new(LoopState::Inactive, "loop", |s| match s.as_ref() {
        "Track" => LoopState::Single,
        "Playlist" => LoopState::All,
        "None" => LoopState::None,
        "" => LoopState::Inactive,
        _ => panic!("Unknown state: {}", s),
    })
});

pub async fn manage_button(path: KeyPath, config: ButtonConfig, mut events: broadcast::Receiver<(KeyPath, KeyEvent)>, command: ButtonCommand) {
    let mut is_active = false;
    let mut state = STATE_WATCHER_PLAYING.subscribe_to_state();

    let command = match command {
        ButtonCommand::PlayPause => "play-pause",
        ButtonCommand::Play => "play",
        ButtonCommand::Pause => "pause",
        ButtonCommand::Previous => "previous",
        ButtonCommand::Next => "next",
    };

    loop {
        tokio::select! {
            result = state.recv() => {
                match result {
                    Err(RecvError::Closed) => { result.unwrap(); },
                    Err(RecvError::Lagged(_)) => { /* mission failed, we'll get 'em next time */ },
                    Ok(state) => {
                        is_active = state != ButtonState::Inactive;

                        send_command(HandlerCommand::SetKeyStyle {
                            path: path.clone(),
                            value: config.style.get(&state).cloned()
                        });
                    }
                }
            }

            Ok((p, event)) = events.recv() => {
                if p != path {
                    continue
                }

                if event == KeyEvent::Press && is_active {
                    let status = std::process::Command::new("playerctl").arg(command).status().unwrap();

                    if !status.success() {
                        log::error!("`playerctl {}` failed with exit code {}", command, status)
                    }
                }
            }
        }
    }
}

pub async fn manage_shuffle(path: KeyPath, config: ShuffleConfig, mut events: broadcast::Receiver<(KeyPath, KeyEvent)>) {
    let mut state = STATE_WATCHER_SHUFFLE.subscribe_to_state();

    loop {
        tokio::select! {
            result = state.recv() => {
                match result {
                    Err(RecvError::Closed) => { result.unwrap(); },
                    Err(RecvError::Lagged(_)) => { /* mission failed, we'll get 'em next time */ },
                    Ok(state) => {
                        send_command(HandlerCommand::SetKeyStyle {
                            path: path.clone(),
                            value: config.style.get(&state).cloned()
                        })
                    }
                }
            }

            Ok((p, event)) = events.recv() => {
                if p != path {
                    continue
                }

                if event == KeyEvent::Press {
                    let current = *STATE_WATCHER_SHUFFLE.state.read().unwrap();
                    let new = match current {
                        ShuffleState::Inactive | ShuffleState::Off => "On",
                        ShuffleState::On => "Off",
                    };

                    let status = std::process::Command::new("playerctl").args(["shuffle", new]).status().unwrap();

                    if !status.success() {
                        log::error!("`playerctl shuffle {}` failed with exit code {}", new, status)
                    }
                }
            }
        }
    }
}

pub async fn manage_loop(path: KeyPath, config: LoopConfig, mut events: broadcast::Receiver<(KeyPath, KeyEvent)>) {
    let mut state = STATE_WATCHER_LOOP.subscribe_to_state();

    loop {
        tokio::select! {
            result = state.recv() => {
                match result {
                    Err(RecvError::Closed) => { result.unwrap(); },
                    Err(RecvError::Lagged(_)) => { /* mission failed, we'll get 'em next time */ },
                    Ok(state) => {
                        send_command(HandlerCommand::SetKeyStyle {
                            path: path.clone(),
                            value: config.style.get(&state).cloned()
                        })
                    }
                }
            }

            Ok((p, event)) = events.recv() => {
                if p != path {
                    continue
                }

                if event == KeyEvent::Press {
                    let current = *STATE_WATCHER_LOOP.state.read().unwrap();
                    let new = match current {
                        LoopState::Inactive | LoopState::None => "Playlist",
                        LoopState::All => "Track",
                        LoopState::Single => "None",
                    };

                    let status = std::process::Command::new("playerctl").args(["loop", new]).status().unwrap();

                    if !status.success() {
                        log::error!("`playerctl loop {}` failed with exit code {}", new, status)
                    }
                }
            }
        }
    }
}

pub struct Handler {
    events_sender: broadcast::Sender<(KeyPath, KeyEvent)>,
    #[allow(unused)]
    runtime: tokio::runtime::Runtime,
}

impl Handler {
    pub fn new(data: InitialHandlerMessage<KeyConfig, ()>) -> Result<Self, HandlerInitializationError> {
        let (events_sender, _) = broadcast::channel::<(KeyPath, KeyEvent)>(5);

        let runtime = tokio::runtime::Builder::new_multi_thread().worker_threads(1).build().unwrap();

        for (path, config) in data.key_configs {
            let events = events_sender.subscribe();

            match config {
                KeyConfig::Loop(config) => runtime.spawn(manage_loop(path, config, events)),
                KeyConfig::Shuffle(config) => runtime.spawn(manage_shuffle(path, config, events)),
                KeyConfig::Previous(config) => runtime.spawn(manage_button(path, config, events, ButtonCommand::Previous)),
                KeyConfig::Next(config) => runtime.spawn(manage_button(path, config, events, ButtonCommand::Next)),
                KeyConfig::Pause(config) => runtime.spawn(manage_button(path, config, events, ButtonCommand::Pause)),
                KeyConfig::Play(config) => runtime.spawn(manage_button(path, config, events, ButtonCommand::Play)),
                KeyConfig::PlayPause(config) => runtime.spawn(manage_button(path, config, events, ButtonCommand::PlayPause)),
            };
        }

        Ok(Handler { events_sender, runtime })
    }
}

impl DecksterHandler for Handler {
    fn handle(&mut self, event: HandlerEvent) {
        if let HandlerEvent::Key { path, event } = event {
            self.events_sender.send((path, event)).unwrap();
        }
    }
}
