use std::str::FromStr;
use std::time::Duration;

use rumqttc::{Event, Incoming, LastWill, MqttOptions, QoS};
use tokio::sync::broadcast;

use crate::coordinator::io_worker::CoordinatorCommand;
use deckster_shared::handler_communication::{HandlerCommand, HandlerEvent};
use deckster_shared::path::{KeyPath, KeyPosition, KnobPath, KnobPosition};
use deckster_shared::style::{KeyStyle, KnobStyle};

use crate::model::mqtt::{HandlerHostsConfig, MqttConfig};

pub async fn start_mqtt_client(
    config: &MqttConfig,
    handler_hosts_config: &HandlerHostsConfig,
    coordinator_commands_sender: flume::Sender<CoordinatorCommand>,
    handler_commands_sender: flume::Sender<HandlerCommand>,
    mut events_receiver: broadcast::Receiver<HandlerEvent>,
) {
    let topic_prefix = config.topic_prefix.to_owned();
    let config_topic = format!("{topic_prefix}/config");

    let handler_hosts_config_json = serde_json::to_string(handler_hosts_config).unwrap().into_bytes();

    let mut options = MqttOptions::new(&config.client_id, &config.host, config.port);
    options.set_keep_alive(Duration::from_secs(3));
    options.set_clean_session(true);

    options.set_last_will(LastWill::new(
        config_topic.clone(),
        serde_json::to_vec(&Option::<HandlerHostsConfig>::None).unwrap(),
        QoS::ExactlyOnce,
        true,
    ));

    if let Some(credentials) = &config.credentials {
        options.set_credentials(&credentials.username, &credentials.password);
    }

    let (client, mut event_loop) = rumqttc::AsyncClient::new(options, 32);

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();
        let client = client.clone();

        async move {
            while let Ok(event) = events_receiver.recv().await {
                match event {
                    HandlerEvent::Stop => {
                        unreachable!("Not expected to ever be sent on the coordinator side")
                    }
                    HandlerEvent::Key { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/keys/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                    HandlerEvent::Knob { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/knobs/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                }
            }
        }
    });

    let run_id = handler_hosts_config.run_id.clone().into_boxed_bytes();

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();
        let client = client.clone();

        async move {
            let mut is_connected = false;
            let mut is_first_try = true;

            loop {
                match event_loop.poll().await {
                    Err(error) => {
                        if is_first_try {
                            log::error!("MQTT connection failed (retrying in the background): {error}");
                            is_first_try = false;
                        } else if is_connected {
                            log::error!("MQTT connection lost: {error}");

                            coordinator_commands_sender.send(CoordinatorCommand::SetAllRemoteHostsInactive).unwrap();
                        }

                        is_connected = false;
                        tokio::time::sleep(Duration::from_secs(2)).await;
                    }
                    Ok(Event::Incoming(event)) => match event {
                        Incoming::ConnAck(_) => {
                            log::info!("MQTT connection established.");
                            is_connected = true;
                            is_first_try = false;

                            client.subscribe(format!("{topic_prefix}/remote_hosts/+"), QoS::AtLeastOnce).await.unwrap();
                            client.subscribe(format!("{topic_prefix}/keys/+/+/style"), QoS::ExactlyOnce).await.unwrap();
                            client.subscribe(format!("{topic_prefix}/knobs/+/+/style"), QoS::ExactlyOnce).await.unwrap();
                            client.subscribe(format!("{topic_prefix}/knobs/+/+/value"), QoS::ExactlyOnce).await.unwrap();

                            client
                                .publish(config_topic.clone(), QoS::ExactlyOnce, true, handler_hosts_config_json.clone())
                                .await
                                .unwrap()
                        }
                        Incoming::Publish(event) => {
                            let segments = event.topic.strip_prefix(&topic_prefix).unwrap().split('/').skip(1).collect::<Vec<&str>>();

                            match segments[0] {
                                "remote_hosts" => {
                                    let host_id = segments[1].to_owned().into_boxed_str();

                                    coordinator_commands_sender
                                        .send(CoordinatorCommand::SetRemoteHostIsActive {
                                            host_id,
                                            is_active: event.payload == run_id.as_ref(),
                                        })
                                        .unwrap();
                                }
                                "keys" => {
                                    let page_id = segments[1];
                                    let position = segments[2];
                                    let property = segments[3];

                                    if property == "style" {
                                        if let Ok(position) = KeyPosition::from_str(position) {
                                            if let Ok(value) = serde_json::from_slice::<Option<KeyStyle>>(&event.payload) {
                                                handler_commands_sender
                                                    .send_async(HandlerCommand::SetKeyStyle {
                                                        path: KeyPath {
                                                            page_id: page_id.to_owned(),
                                                            position,
                                                        },
                                                        value,
                                                    })
                                                    .await
                                                    .unwrap();
                                            } else {
                                                log::error!("Could not deserialize the latest key style object from {}", event.topic);
                                            }
                                        } else {
                                            log::warn!("Invalid key position in topic name: {}", event.topic);
                                        }
                                    }
                                }
                                "knobs" => {
                                    let page_id = segments[1];
                                    let position = segments[2];
                                    let property = segments[3];

                                    if let Ok(position) = KnobPosition::from_str(position) {
                                        match property {
                                            "style" => {
                                                if let Ok(value) = serde_json::from_slice::<Option<KnobStyle>>(&event.payload) {
                                                    handler_commands_sender
                                                        .send_async(HandlerCommand::SetKnobStyle {
                                                            path: KnobPath {
                                                                page_id: page_id.to_owned(),
                                                                position,
                                                            },
                                                            value,
                                                        })
                                                        .await
                                                        .unwrap();
                                                }
                                            }
                                            "value" => {
                                                if let Ok(value) = serde_json::from_slice::<Option<f32>>(&event.payload) {
                                                    handler_commands_sender
                                                        .send_async(HandlerCommand::SetKnobValue {
                                                            path: KnobPath {
                                                                page_id: page_id.to_owned(),
                                                                position,
                                                            },
                                                            value,
                                                        })
                                                        .await
                                                        .unwrap();
                                                }
                                            }
                                            _ => {}
                                        }
                                    } else {
                                        log::warn!("Invalid knob position in topic name: {}", event.topic);
                                    }
                                }
                                _ => {}
                            };
                        }
                        _ => {}
                    },
                    _ => {}
                }
            }
        }
    });
}
