use std::cmp::max;
use std::collections::HashMap;
use std::path::{Path, PathBuf};
use std::sync::Arc;

use color_eyre::eyre::{eyre, ContextCompat, WrapErr};
use color_eyre::Result;
use resvg::usvg::fontdb;
use resvg::usvg::tiny_skia_path::IntSize;
use resvg::usvg::{TextRendering, TreeParsing, TreeTextToPath};
use tiny_skia::{BlendMode, FilterQuality, Pixmap, PixmapPaint, Transform};

use deckster_shared::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use deckster_shared::image_filter::ImageFilter;

use crate::model::coordinator_config::{IconFormat, IconPack};
use crate::util::UnwrapExt;

mod destructive_filter;

#[derive(Debug)]
pub struct IconManager {
    config_directory: PathBuf,
    icon_packs_by_id: Arc<HashMap<String, IconPack>>,
    dpi: f32,
    fonts_db: fontdb::Database,
}

#[derive(Debug)]
struct LoadedIcon {
    pixmap: Pixmap,
    effective_filter: ImageFilter,
}

impl IconManager {
    pub fn new(config_directory: PathBuf, icon_packs_by_id: Arc<HashMap<String, IconPack>>, dpi: f32) -> IconManager {
        let mut fonts_db = fontdb::Database::new();
        fonts_db.load_system_fonts();

        IconManager {
            config_directory,
            icon_packs_by_id,
            dpi,
            fonts_db,
        }
    }

    // TODO: Add caching
    fn load_icon(&self, descriptor: &IconDescriptor) -> Result<LoadedIcon> {
        let icon_pack = if let IconDescriptorSource::IconPack { pack_id, .. } = &descriptor.source {
            Some(
                self.icon_packs_by_id
                    .get(pack_id)
                    .wrap_err_with(|| format!("Unknown icon pack: @{}", pack_id))?,
            )
        } else {
            None
        };

        let icon_pack_filter = icon_pack.and_then(|p| p.global_filter.as_ref());

        let mut effective_filter = icon_pack_filter
            .map(|f| descriptor.filter.merge_over(f))
            .unwrap_or_else(|| descriptor.filter.clone());

        let (original_image, original_image_scale) = read_image_and_get_scale(
            &self.config_directory,
            self.dpi,
            &self.fonts_db,
            &descriptor.source,
            icon_pack,
            effective_filter.transform.scale,
        )?;

        let pixmap = destructive_filter::apply(&original_image, &effective_filter.destructive)?;

        effective_filter.transform.scale /= original_image_scale;

        Ok(LoadedIcon { pixmap, effective_filter })
    }

    pub fn render_icon_in(&self, pixmap: &mut Pixmap, descriptor: &IconDescriptor) -> Result<()> {
        let icon = self.load_icon(descriptor)?;
        let scale = icon.effective_filter.transform.scale;

        let scaled_size = IntSize::from_wh(icon.pixmap.width(), icon.pixmap.height())
            .expect("width and height are not zero")
            .scale_by(scale)
            .unwrap_todo();

        pixmap.draw_pixmap(
            (((pixmap.width() as i32 - scaled_size.width() as i32) / 2) as f32 / scale).round() as i32,
            (((pixmap.height() as i32 - scaled_size.height() as i32) / 2) as f32 / scale).round() as i32,
            icon.pixmap.as_ref(),
            &PixmapPaint {
                opacity: icon.effective_filter.transform.alpha,
                blend_mode: BlendMode::SourceOver,
                quality: FilterQuality::Bicubic,
            },
            Transform::from_scale(scale, scale).post_rotate_at(
                (icon.effective_filter.transform.clockwise_quarter_rotations as f32) * 90.0,
                pixmap.width() as f32 / 2.0,
                pixmap.height() as f32 / 2.0,
            ),
            None,
        );

        Ok(())
    }
}

// Panics when `source` is `IconPack` and `icon_pack` is `None`.
fn read_image_and_get_scale(
    config_directory: &Path,
    dpi: f32,
    fonts_db: &fontdb::Database,
    source: &IconDescriptorSource,
    icon_pack: Option<&IconPack>,
    preferred_scale: f32,
) -> Result<(Pixmap, f32)> {
    let path = match source {
        IconDescriptorSource::Path(path) => path.clone(),
        IconDescriptorSource::IconPack { icon_id, .. } => {
            let icon_pack = icon_pack.unwrap();
            let extension = match icon_pack.format {
                IconFormat::Png => "png",
                IconFormat::Svg => "svg",
            };

            config_directory.join(&icon_pack.path).join(icon_id.to_owned() + "." + extension)
        }
    };

    Ok(match path.extension() {
        None => return Err(eyre!("Invalid icon path: {:?}", path)),
        Some(extension) => match extension.to_string_lossy().as_ref() {
            "png" => (
                Pixmap::load_png(&path).wrap_err_with(|| format!("Failed to open or decode the PNG file at {}", path.to_string_lossy()))?,
                1.0,
            ),
            "svg" => (
                read_image_from_svg(&path, dpi, fonts_db, preferred_scale)
                    .wrap_err_with(|| format!("Failed to open or decode the SVG file at {}", path.to_string_lossy()))?,
                preferred_scale,
            ),
            extension => return Err(eyre!("Invalid file extension, only *.png and *.svg are allowed: {}", extension)),
        },
    })
}

fn read_image_from_svg(path: &Path, dpi: f32, font_db: &fontdb::Database, scale: f32) -> Result<Pixmap> {
    let raw_data = std::fs::read(path)?;

    let tree = {
        let mut tree = resvg::usvg::Tree::from_data(
            &raw_data,
            &resvg::usvg::Options {
                dpi,
                font_family: "Inter".to_owned(),
                font_size: 11.0,
                text_rendering: TextRendering::OptimizeLegibility,
                ..resvg::usvg::Options::default()
            },
        )?;

        tree.convert_text(font_db);

        resvg::Tree::from_usvg(&tree)
    };

    let size = tree.size.to_int_size();
    let mut pixmap = Pixmap::new(
        max(1, (size.width() as f32 * scale).ceil() as u32),
        max(1, (size.height() as f32 * scale).ceil() as u32),
    )
    .expect("width and height can never be zero.");

    tree.render(Transform::from_scale(scale, scale), &mut pixmap.as_mut());

    Ok(pixmap)
}
