use std::fmt::{Display, Formatter};
use std::str::FromStr;

use derive_more::{Deref, From, Into};
use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;
use tiny_skia::IntRect;

#[derive(Debug, Eq, PartialEq, Hash, Clone, SerializeDisplay, DeserializeFromStr)]
pub struct UIntVec2 {
    pub x: u32,
    pub y: u32,
}

#[derive(Debug, Error)]
#[error("The input value does not match the required format of <x> and <y> separated by an 'x'")]
pub struct UIntVec2FromStrError {}

impl FromStr for UIntVec2 {
    type Err = UIntVec2FromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let values = s.split_once('x');

        if let Some((x, y)) = values {
            if let Ok(x) = u32::from_str(x) {
                if let Ok(y) = u32::from_str(y) {
                    return Ok(UIntVec2 { x, y });
                }
            }
        }

        Err(UIntVec2FromStrError {})
    }
}

impl Display for UIntVec2 {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}x{}", self.x, self.y))
    }
}

#[derive(Debug, Error)]
pub enum IntRectWrapperFromStrErr {
    #[error("The input value does not match the required format.")]
    ParsingFailed,

    #[error("The right bottom coordinates are not strictly higher than the left top coordinates.")]
    ConstraintsFailed,
}

#[derive(Debug, Copy, Clone, PartialEq, Deref, From, Into, SerializeDisplay, DeserializeFromStr)]
#[repr(transparent)]
pub struct IntRectWrapper(IntRect);

impl FromStr for IntRectWrapper {
    type Err = IntRectWrapperFromStrErr;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (pairs, is_corner_points_mode) = if let Some(pairs) = s.split_once('+') {
            (pairs, false)
        } else if let Some(pairs) = s.split_once('-') {
            (pairs, true)
        } else {
            return Err(IntRectWrapperFromStrErr::ParsingFailed);
        };

        let first_vec = UIntVec2::from_str(pairs.0).map_err(|_| IntRectWrapperFromStrErr::ParsingFailed)?;
        let second_vec = UIntVec2::from_str(pairs.1).map_err(|_| IntRectWrapperFromStrErr::ParsingFailed)?;

        Ok(IntRectWrapper(if is_corner_points_mode {
            IntRect::from_ltrb(first_vec.x as i32, first_vec.y as i32, second_vec.x as i32, second_vec.y as i32)
                .ok_or(IntRectWrapperFromStrErr::ConstraintsFailed)?
        } else {
            IntRect::from_xywh(first_vec.x as i32, first_vec.y as i32, second_vec.x, second_vec.y).ok_or(IntRectWrapperFromStrErr::ConstraintsFailed)?
        }))
    }
}

impl Display for IntRectWrapper {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}x{}+{}x{}", self.x(), self.y(), self.width(), self.height()))
    }
}
